#!/usr/bin/python

"""
netsim DPOR

Implements dynamic partial order reduction using vector clocks

This is basically a copy of netsim.py with some stuff changed.
"""

import hashlib
import struct
import zlib


class Adler32Hash(object):
    """hashlib interface to adler32. Faster than sha1, but not by much."""

    __slots__ = ("value")

    def __init__(self, initial_string):
        self.value = zlib.adler32(initial_string)

    def digest(self):
        return struct.pack("L", self.value)

    def update(self, string):
        self.value = zlib.adler32(string, self.value)

HASH_TYPE = hashlib.sha1
#~ HASH_TYPE = Adler32Hash


class Message(object):
    """Contains data along with header information."""

    __slots__ = ("source", "destination", "id", "data", "vector_clock")

    def __init__(self, id, source, destination, data, vector_clock):
        assert source >= 0
        assert destination >= 0
        assert source != destination
        assert len(vector_clock) >= 1

        self.source = source
        self.destination = destination
        self.id = id
        self.data = data
        self.vector_clock = vector_clock

    def __eq__(self, other):
        equal = self.id == other.id
        # Don't assert data is equal: it might be unserialized messages that don't implement __eq__
        assert (not equal) or (self.source == other.source and
                self.destination == other.destination and self.vector_clock == other.vector_clock)
        return equal

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "%d -> %d: %s" % (self.source, self.destination, str(self.data))


def vectorClockMergeIncrement(v1, v2, index):
    """Returns a vector clock that is the result of merging v1 and v2."""

    assert len(v1) == len(v2)
    for i, v in enumerate(v1):
        if v2[i] > v:
            v1[i] = v2[i]
    v1[index] += 1
    

def vectorClockIncrement(v, index):
    v_list = list(v)
    v_list[index] += 1
    return tuple(v_list)


def vectorClockMerge(v1, v2):
    """Returns a vector clock that is the result of merging v1 and v2."""

    assert len(v1) == len(v2)
    return tuple([max(values) for values in zip(v1, v2)])


def vectorClockCompare(v1, v2):
    """Returns -1 if v1 happens-before v2, 1 if v2 happens-before v1, 0 if they are concurrent."""

    assert len(v1) == len(v2)

    foundLessThan = False
    foundGreaterThan = False
    for a, b in zip(v1, v2):
        if a < b:
            foundLessThan = True
        elif a > b:
            foundGreaterThan = True
        else:
            assert a == b

    if foundLessThan and not foundGreaterThan:
        # All in v1 are <= v2, and at least one is <: v1 happens-before v2
        return -1
    if foundGreaterThan and not foundLessThan:
        # All in v1 are >= v2, and at least one is >: v2 happens-before v1
        return 1
    else:
        # Concurrent
        assert foundGreaterThan and foundLessThan or tuple(v1) == tuple(v2)
        return 0


class NodeState(object):
    """Encapsulates the current state of a node.
    
    The history of events is represented by hashing IDs. Each time a message is
    received, the id of the message is hashed with the current history hash.
    Before a message is sent, the hishory ID is hashed with itself. The
    message's id is equal to the state that generated it.
    """

    __slots__ = ("state_machine", "history_hash", "hash_cache", "message_round_count",
            "node_id", "vector_clock")

    def __init__(self, state_machine):
        assert state_machine is not None

        self.state_machine = state_machine
        self.history_hash = None
        self.hash_cache = None
        self.message_round_count = 0
        self.node_id = None
        self.vector_clock = None

    def historyHash(self):
        """Returns the hash of this node's sequence of operations."""
        return self.hash_cache

    def start(self, system, node_id, num_nodes):
        """Notifies the state machine that the simulation is beginning. node_id must be unique so
        node history ids are unique."""

        assert self.node_id is None
        assert self.history_hash is None
        assert 0 <= node_id and node_id < num_nodes
        self.node_id = node_id
        self.vector_clock = [0,] * num_nodes
        self.vector_clock[self.node_id] += 1

        # Generate the initial id (hopefully unique)
        self.history_hash = HASH_TYPE(struct.pack("L", self.node_id))
        self.hash_cache = self.history_hash.digest()

        self.state_machine.start(system, self.node_id)

    def messageReceived(self, system, message):
        """Deliver message to the state machine in this node."""

        # Update the history hash and clock then deliver the message
        self.history_hash.update(message.id)
        self.hash_cache = self.history_hash.digest()
        self.message_round_count = 0

        vectorClockMergeIncrement(self.vector_clock, message.vector_clock, self.node_id)
        #~ self.vector_clock = vectorClockMerge(self.vector_clock, message.vector_clock)
        #~ self.vector_clock = vectorClockIncrement(self.vector_clock, self.node_id)

        self.state_machine.messageReceived(system, message.source, message.data)

    def getSentMessageId(self):
        """Returns the id of a message to be sent."""
        message_id = self.historyHash()
        
        # Message id is the node's state XOR sent count
        if self.message_round_count > 0:
            last_byte = ord(message_id[-1])
            assert self.message_round_count < 256
            last_byte ^= self.message_round_count
            message_id = message_id[:-1] + chr(last_byte)
        self.message_round_count += 1

        return message_id

    def getSentVectorClock(self):
        #~ self.vector_clock[self.node_id] += 1
        #~ self.vector_clock = vectorClockIncrement(self.vector_clock, self.node_id)
        self.vector_clock[self.node_id] += 1
        return tuple(self.vector_clock)


class VisitedStates(object):
    __slots__ = ("delivered", "to_visit")

    def __init__(self, num_nodes):
        self.delivered = [set() for _ in xrange(num_nodes)]
        self.to_visit = set()

    def can_deliver(self, destination, event_id):
        return event_id not in self.delivered[destination]

    def deliver(self, destination, event_id):
        """Returns a new VisitedStates object that represents the state after delivering an event
        to destination. If we are not permitted to deliver event_id, returns None."""

        assert self.can_deliver(destination, event_id)
        self.to_visit.discard(event_id)
        self.delivered[destination].add(event_id)

        # Clone this object
        result = VisitedStates(len(self.delivered))
        for i, s in enumerate(self.delivered):
            result.delivered[i] = s.copy()

        # Replace the destination with the empty set as all messages are now
        # permitted
        result.delivered[destination] = set()

        return result

    def queue(self, event_id):
        self.to_visit.add(event_id)

    def next(self):
        if len(self.to_visit) > 0:
            return iter(self.to_visit).next()
        return None


class System(object):
    """Provides a framework for simulating asynchronous networks."""

    __slots__ = ("nodes", "events", "current_node")

    def __init__(self):
        self.nodes = []
        self.events = {}
        self.current_node = None

    def addNode(self, node):
        assert node not in [node_state.state_machine for node_state in self.nodes]
        node_id = len(self.nodes)
        self.nodes.append(NodeState(node))
        return node_id

    def startStateMachines(self):
        """Initializes all nodes by calling their start methods."""
        assert len(self.nodes) > 0
        assert len(self.events) == 0
        for node_id, node in enumerate(self.nodes):
            self.current_node = node_id
            node.start(self, node_id, len(self.nodes))
            self.current_node = None

        #~ for i in [1, 2, 0, 1, 0, 1, 2]:
            #~ print self.events.values()[i]
            #~ self.executeChoice(self.events.keys()[i])

    def executeChoice(self, event_id):
        # Remove the event from the queue
        event = self.events[event_id]
        del self.events[event_id]
        self.current_node = event.destination
        #~ print choice, event.source, "->", event.destination, event.data, self.current_node
        self.nodes[self.current_node].messageReceived(self, event)
        #~ def summary(hash):
            #~ bytes = hash.digest()[:4]
            #~ return "".join(["%02x" % ord(x) for x in bytes])
        #~ print [summary(n.history_hash) for n in self.nodes]
        self.current_node = None

    def simulateDeterministic(self):
        assert len(self.nodes) > 0
        assert len(self.events) == 0

        # Simulate until there are no more events
        self.startStateMachines()
        while len(self.events) > 0:
            self.executeChoice(self.events.iterkeys().next())

    def resimulate(self, initializer, choices):
        """Simulates the events represented by choices."""
        # Call the initializer to create the initial set of nodes
        # TODO: Test determinism?
        System.__init__(self)
        initializer(self)
        self.startStateMachines()

        # Replay history
        for choice in choices:
            self.executeChoice(choice)
    
    def sendMessage(self, destination, message):
        assert self.current_node is not None
        assert destination is not None
        assert destination < len(self.nodes)

        message_id = self.nodes[self.current_node].getSentMessageId()
        vector_clock = self.nodes[self.current_node].getSentVectorClock()
        assert message_id not in self.events
        self.events[message_id] = Message(message_id, self.current_node, destination, message, vector_clock)


def listIsPrefix(l, prefix):
    """Returns true if prefix is a prefix for l."""

    if len(l) < len(prefix):
        return False

    # Check using Python slice notation
    return l[:len(prefix)] == prefix


class EventHistory(object):
    __slots__ = ("event", "vector_clock")

    def __init__(self, event, vector_clock):
        self.event = event
        self.vector_clock = vector_clock
        assert vectorClockCompare(self.event.vector_clock, vector_clock) == -1


class SimulatedSystem(object):
    __slots__ = (
            "initializer", "system", "event_history")

    def __init__(self, initializer):
        # Initialize a system to get the number of nodes
        self.initializer = initializer
        self.reset()

    def numNodes(self):
        return len(self.system.nodes)

    def reset(self):
        """Resets the system state."""
        self.system = System()
        self.initializer(self.system)
        self.system.startStateMachines()

        self.event_history = []

    def resimulate(self, choices):
        self.reset()
        # Execute forward
        for choice in choices:
            self.executeChoice(choice)
        #~ assert self.choices == choices

    def executeChoice(self, event_id):
        event = self.system.events[event_id]

        # Get the vector clock for the receive event
        # HUGE HACK: Fix this
        vector_clock = vectorClockMerge(self.system.nodes[event.destination].vector_clock,
                event.vector_clock)
        vector_clock = vectorClockIncrement(vector_clock, event.destination)

        self.system.executeChoice(event_id)

        # Update the history for dependency tracking
        history = EventHistory(event, vector_clock)
        self.event_history.append(history)

    def stateMachines(self):
        return [n.state_machine for n in self.system.nodes]


#~ def simulateDPOR(initializer, validator):
def simulateAll(initializer, validator):
    """Simulates all possible outcomes using dynamic partial order reduction.
    
    initializer: called to create the initial state in a deterministic fashion
    validator: returns true if the final state is okay."""

    system = SimulatedSystem(initializer)

    # sequences of choices that fail validation
    error_choices = []

    choices = []
    history_stack = [VisitedStates(system.numNodes())]

    execute_count = 0
    while len(history_stack) > 0:
        execute_count += 1

        # Replay history: we end up at a new state
        #~ system.resimulate(choices)
        #~ print choices

        # Make new choices
        while len(system.system.events) > 0:
            # Pick an event we can deliver
            event = None
            for e in system.system.events.itervalues():
                if history_stack[-1].can_deliver(e.destination, e.id):
                    event = e
                    break
            if event is None:
                break
            assert history_stack[-1].can_deliver(event.destination, event.id)

            # execute the choice
            choices.append(event.id)
            system.executeChoice(event.id)
            assert len(choices) == len(system.event_history)

            # Mark as visited
            new_history = history_stack[-1].deliver(event.destination, event.id)
            history_stack.append(new_history)

            # DPOR: previous actions to find conflicting events caused by concurrent messages
            for i in xrange(len(system.event_history)-2, -1, -1):
                h = system.event_history[i]

                # do the messages conflict? if not, we don't care
                if h.event.destination != event.destination:
                    continue

                # is it possible to deliver this message before receiving the other?
                # (concurrent or event send happens-before h receive)
                diff = vectorClockCompare(h.vector_clock, event.vector_clock)
                if diff == -1:
                    # if receive h happens before send event, then h *caused* event. We can't reorder.
                    # even better, any previous conflicting event will also be causal
                    #~ for j, h in enumerate(system.event_history[:i]):
                        #~ assert h.event.destination != event.destination or vectorClockCompare(
                                #~ h.vector_clock, event.vector_clock) == -1
                    break
                else:
                    # The receive can be either be CONCURRENT or AFTER the send
                    #~ assert diff == 0
                    pass

                #~ print i, "conflicts -concurrent-with-", len(choices)-1, h.event.data, event.data
                # Check every event's vector clock
                next_event = event
                for index in xrange(i+1, len(choices)-1):
                    #~ print "  dependent?", index, system.event_history[index].vector_clock, event.vector_clock
                    if vectorClockCompare(
                            system.event_history[index].vector_clock, event.vector_clock) == -1:
                        # This is a dependent event:
                        next_event = system.event_history[index].event
                        break

                if history_stack[i].can_deliver(next_event.destination, next_event.id):
                    history_stack[i].queue(next_event.id)

        # This if statement is false if we aborted due to duplicate history
        if len(system.system.events) == 0:
            # We have reached the end of execution: validate
            state_machines = system.stateMachines()
            if not validator(state_machines):
                print "error:", choices
                error_choices.append(choices)

        # Pop off choices until we have more choices to make
        while len(history_stack) > 0 and history_stack[-1].next() is None:
            if len(choices) == 0:
                choices = None
            else:
                choices.pop()
            assert history_stack[-1].next() is None
            history_stack.pop()
        assert len(history_stack) == 0 or history_stack[-1].next() is not None

        if choices is not None:
            next_event_id = history_stack[-1].next()
            system.resimulate(choices)

            choices.append(next_event_id)
            destination = system.system.events[next_event_id].destination
            history_stack.append(history_stack[-1].deliver(destination, next_event_id))
            system.executeChoice(next_event_id)

    # We have exhausted the universe of possible orderings
    print execute_count
    return error_choices
