#!/usr/bin/python

import unittest

import paxos


class MessageTest(unittest.TestCase):
    def testPrepare(self):
        self.assertRaises(AssertionError, paxos.PrepareMessage, None, 0)
        self.assertRaises(AssertionError, paxos.PrepareMessage, 0, None)
        prepare1 = paxos.PrepareMessage(0, 0)
        prepare2 = paxos.PrepareMessage(0, 1)
        prepare3 = paxos.PrepareMessage(1, 0)

        assert prepare1.proposal_number < prepare2.proposal_number
        assert prepare1.proposal_number < prepare3.proposal_number
        assert prepare3.proposal_number < prepare2.proposal_number

    def testAccept(self):
        self.assertRaises(AssertionError, paxos.AcceptMessage, None, 0, "foo")
        self.assertRaises(AssertionError, paxos.AcceptMessage, 0, None, "foo")
        self.assertRaises(AssertionError, paxos.AcceptMessage, 0, 0, None)
        accept1 = paxos.AcceptMessage(0, 0, "foo")
        accept2 = paxos.AcceptMessage(0, 1, "foo")
        accept3 = paxos.AcceptMessage(1, 0, "foo")

        assert accept1.proposal_number < accept2.proposal_number
        assert accept1.proposal_number < accept3.proposal_number
        assert accept3.proposal_number < accept2.proposal_number


class MockSystem(object):
    def __init__(self):
        self.sent = []

    def sendMessage(self, destination, message):
        self.sent.append((destination, message))

    def getSingleSent(self, destination):
        assert len(self.sent) == 1
        assert destination == self.sent[0][0]
        message = self.sent[0][1]
        self.sent = []
        return message


class AcceptorTest(unittest.TestCase):
    def setUp(self):
        self.acceptor = paxos.Acceptor()
        self.system = MockSystem()

    def testPrepare(self):
        self.acceptor.messageReceived(self.system, 0, paxos.PrepareMessage(0, 1))
        m = self.system.getSingleSent(0)
        self.assertEquals((1, 0), m.proposal_number)
        self.assertEquals(True, m.prepared)
        self.assertEquals(None, m.value)
        self.assertEquals((1, 0), self.acceptor.proposal_number)

        # Lower prepare = reject
        self.acceptor.messageReceived(self.system, 1, paxos.PrepareMessage(1, 0))
        m = self.system.getSingleSent(1)
        self.assertEquals((1, 0), m.proposal_number)
        self.assertEquals(False, m.prepared)
        self.assertEquals((1, 0), self.acceptor.proposal_number)

        # Higher prepare = accept again
        self.acceptor.messageReceived(self.system, 2, paxos.PrepareMessage(2, 1))
        m = self.system.getSingleSent(2)
        self.assertEquals((1, 2), m.proposal_number)
        self.assertEquals(True, m.prepared)
        self.assertEquals((1, 2), self.acceptor.proposal_number)

    def testAccept(self):
        # Accept an initial value
        self.acceptor.messageReceived(self.system, 0, paxos.AcceptMessage(0, 2, "v"))
        m = self.system.getSingleSent(0)
        self.assertEquals((2, 0), m.proposal_number)
        self.assertEquals(True, m.accepted)
        self.assertEquals("v", self.acceptor.value)

        # Accept a smaller sequence number: rejected
        self.acceptor.messageReceived(self.system, 1, paxos.AcceptMessage(1, 0, "w"))
        m = self.system.getSingleSent(1)
        self.assertEquals((2, 0), m.proposal_number)
        self.assertEquals(False, m.accepted)
        self.assertEquals("v", self.acceptor.value)

        # Accept a higher sequence number: accepted
        self.acceptor.messageReceived(self.system, 1, paxos.AcceptMessage(1, 3, "w"))
        m = self.system.getSingleSent(1)
        self.assertEquals((3, 1), m.proposal_number)
        self.assertEquals(True, m.accepted)
        self.assertEquals("w", self.acceptor.value)

    def testAcceptPrepare(self):
        # Accept a value
        self.acceptor.messageReceived(self.system, 0, paxos.AcceptMessage(0, 2, "v"))
        m = self.system.getSingleSent(0)
        self.assertEquals((2, 0), m.proposal_number)
        self.assertEquals(True, m.accepted)
        self.assertEquals("v", self.acceptor.value)

        # Prepare higher: okay, but with the old value
        self.acceptor.messageReceived(self.system, 1, paxos.PrepareMessage(1, 3))
        m = self.system.getSingleSent(1)
        self.assertEquals((3, 1), m.proposal_number)
        self.assertEquals(True, m.prepared)
        self.assertEquals("v", m.value)
        self.assertEquals((3, 1), self.acceptor.proposal_number)


class ProposerTest(unittest.TestCase):
    def setUp(self):
        self.proposer = paxos.Proposer([0, 3], "v")
        self.system = MockSystem()

    def testStart(self):
        self.proposer.start(self.system, 4)
        self.assertEquals(2, len(self.system.sent))
        self.assertEquals(0, self.system.sent[0][0])
        self.assertEquals(3, self.system.sent[1][0])
        self.assertEquals((0, 4), self.system.sent[0][1].proposal_number)
        self.assertEquals(0, self.proposer.sequence)

    def testPrepareResponse(self):
        self.proposer.start(self.system, 4)
        self.system.sent = []

        # First accept = nothing
        self.proposer.messageReceived(self.system, 0, paxos.PrepareResponse((0, 4), True, None))
        self.assertEquals(0, len(self.system.sent))

        # Second accept = accept
        self.proposer.messageReceived(self.system, 3, paxos.PrepareResponse((0, 4), True, None))
        self.assertEquals(2, len(self.system.sent))
        self.assertEquals(0, self.system.sent[0][0])
        self.assertEquals(3, self.system.sent[1][0])
        self.assertEquals((0, 4), self.system.sent[0][1].proposal_number)
        self.assertEquals("v", self.system.sent[0][1].value)

    def testPrepareResponseWithValue(self):
        self.proposer.start(self.system, 4)
        self.system.sent = []

        # First accept with a value = nothing
        self.proposer.messageReceived(self.system, 0, paxos.PrepareResponse((0, 4), True, "w"))
        self.assertEquals(0, len(self.system.sent))

        # Second accept = accept with old value
        self.proposer.messageReceived(self.system, 3, paxos.PrepareResponse((0, 4), True, None))
        self.assertEquals(2, len(self.system.sent))
        self.assertEquals("w", self.system.sent[0][1].value)
        self.assertEquals("w", self.proposer.proposal_value)


if __name__ == "__main__":
    unittest.main()
